#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2023, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Functions useful for string manipulations or encoding.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging  # isort:skip

log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import re
from typing import Any, Iterable, overload
from urllib.parse import quote_plus

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'append_docstring',
    'format_docstring',
    'indent',
    'nice_join',
    'snakify',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

def indent(text: str, n: int = 2, ch: str = " ") -> str:
    ''' Indent all the lines in a given block of text by a specified amount.

    Args:
        text (str) :
            The text to indent

        n (int, optional) :
            The amount to indent each line by (default: 2)

        ch (char, optional) :
            What character to fill the indentation with (default: " ")

    '''
    padding = ch * n
    return "\n".join(padding + line for line in text.split("\n"))


def nice_join(seq: Iterable[str], *, sep: str = ", ", conjunction: str = "or") -> str:
    ''' Join together sequences of strings into English-friendly phrases using
    the conjunction ``or`` when appropriate.

    Args:
        seq (seq[str]) : a sequence of strings to nicely join
        sep (str, optional) : a sequence delimiter to use (default: ", ")
        conjunction (str or None, optional) : a conjunction to use for the last
            two items, or None to reproduce basic join behaviour (default: "or")

    Returns:
        a joined string

    Examples:
        >>> nice_join(["a", "b", "c"])
        'a, b or c'

    '''
    seq = [str(x) for x in seq]

    if len(seq) <= 1 or conjunction is None:
        return sep.join(seq)
    else:
        return f"{sep.join(seq[:-1])} {conjunction} {seq[-1]}"


def snakify(name: str, sep: str = "_") -> str:
    ''' Convert CamelCase to snake_case. '''
    name = re.sub("([A-Z]+)([A-Z][a-z])", rf"\1{sep}\2", name)
    name = re.sub("([a-z\\d])([A-Z])", rf"\1{sep}\2", name)
    return name.lower()

def append_docstring(docstring: str | None, extra: str) -> str | None:
    ''' Safely append to docstrings.

    When Python is executed with the ``-OO`` option, doc strings are removed and
    replaced the value ``None``. This function guards against appending the
    extra content in that case.

    Args:
        docstring (str or None) : The docstring to format, or None
        extra (str): the content to append if docstring is not None

    Returns:
        str or None

    '''
    return None if docstring is None else docstring + extra

@overload
def format_docstring(docstring: None, *args: Any, **kwargs: Any) -> None: ...
@overload
def format_docstring(docstring: str, *args: Any, **kwargs: Any) -> str: ...

def format_docstring(docstring: str | None, *args: Any, **kwargs: Any) -> str | None:
    ''' Safely format docstrings.

    When Python is executed with the ``-OO`` option, doc strings are removed and
    replaced the value ``None``. This function guards against applying the string
    formatting options in that case.

    Args:
        docstring (str or None) : The docstring to format, or ``None``
        args (tuple) : string formatting arguments for the docsring
        kwargs (dict) : string formatting arguments for the docsring

    Returns:
        str or None

    '''
    return None if docstring is None else docstring.format(*args, **kwargs)


def format_url_query_arguments(url: str, arguments: dict[str, str] | None = None) -> str:
    ''' Format a base URL with optional query arguments

    Args:
        url (str) :
            An base URL to append query arguments to
        arguments (dict or None, optional) :
            A mapping of key/value URL query arguments, or None (default: None)

    Returns:
        str

    '''
    if arguments is not None:
        items = (f"{quote_plus(key)}={quote_plus(value)}" for key, value in arguments.items())
        url += "?" + "&".join(items)
    return url

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2023, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
""" Thoroughly document Bokeh options classes.

The ``bokeh-options`` directive will automatically document all the properties
of a Bokeh Options class under a heading of "Keyword Args".

This directive takes the name of a Bokeh Options subclass as the argument, and
its module as an option:

.. code-block:: rest

    .. bokeh-options:: Opts
        :module: bokeh.sphinxext.sample

Examples
--------

For the following definition of ``bokeh.sphinxext.sample.Opts``:

.. code-block:: python

    class Opts(Options):
        ''' This is an Options class '''

        host = String(default="localhost", help="a host to connect to")
        port = Int(default=5890, help="a port to connect to")

the above usage yields the output:

    .. bokeh-options:: Opts
        :module: bokeh.sphinxext.sample

To enable this extension, add `"bokeh.sphinxext.bokeh_options"` to the
extensions list in your Sphinx configuration module.

"""

# -----------------------------------------------------------------------------
# Boilerplate
# -----------------------------------------------------------------------------
from __future__ import annotations

import logging  # isort:skip

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

# Standard library imports
import importlib
import textwrap

# External imports
from docutils.parsers.rst.directives import unchanged
from sphinx.errors import SphinxError

# Bokeh imports
from bokeh.core.property._sphinx import type_link
from bokeh.util.options import Options

# Bokeh imports
from . import PARALLEL_SAFE
from .bokeh_directive import BokehDirective, py_sig_re
from .templates import OPTIONS_DETAIL

# -----------------------------------------------------------------------------
# Globals and constants
# -----------------------------------------------------------------------------

__all__ = (
    "BokehOptionsDirective",
    "setup",
)

# -----------------------------------------------------------------------------
# General API
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Dev API
# -----------------------------------------------------------------------------


class BokehOptionsDirective(BokehDirective):

    has_content = True
    required_arguments = 1
    optional_arguments = 1
    option_spec = {"module": unchanged}

    def run(self):
        sig = " ".join(self.arguments)

        m = py_sig_re.match(sig)
        if m is None:
            raise SphinxError(f"Unable to parse signature for bokeh-options: {sig!r}")
        name_prefix, options_name, arglist, retann = m.groups()

        module_name = self.options["module"]

        try:
            module = importlib.import_module(module_name)
        except ImportError:
            raise SphinxError(f"Unable to generate options reference docs for {options_name}, couldn't import module {module_name}")

        options = getattr(module, options_name, None)
        if options is None:
            raise SphinxError(f"Unable to generate options reference docs: no options {options_name} in module {module_name}")

        if not issubclass(options, Options):
            raise SphinxError(f"Unable to generate options reference docs: {options_name} is not a subclass of Options")

        options_obj = options({})

        opts = []
        for prop_name in sorted(options_obj.properties()):
            descriptor = options_obj.lookup(prop_name)
            opts.append(
                dict(
                    name=prop_name,
                    type=type_link(descriptor.property),
                    default=repr(descriptor.instance_default(options_obj)),
                    doc="" if descriptor.__doc__ is None else textwrap.dedent(descriptor.__doc__.rstrip()),
                ),
            )

        rst_text = OPTIONS_DETAIL.render(opts=opts)

        return self.parse(rst_text, "<bokeh-options>")


def setup(app):
    """ Required Sphinx extension setup function. """
    app.add_directive_to_domain("py", "bokeh-options", BokehOptionsDirective)

    return PARALLEL_SAFE

# -----------------------------------------------------------------------------
# Private API
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Code
# -----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2023, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Represent array expressions to be computed on the client (browser) side
by BokehJS.

Expression models are useful as ``DataSpec`` values when it is desired that
the array values be computed in the browser:

.. code-block:: python

    p.circle(x={'expr': some_expression}, ...)

or using the ``expr`` convenience function:

.. code-block:: python

    from bokeh.core.properties import expr

    p.circle(x=expr(some_expression), ...)

In this case, the values of the ``x`` coordinates will be computed in the
browser by the JavaScript implementation of ``some_expression`` using a
``ColumnDataSource`` as input.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from math import inf

# Bokeh imports
from ..core.enums import Direction
from ..core.has_props import abstract
from ..core.properties import (
    AngleSpec,
    AnyRef,
    Bool,
    Dict,
    Enum,
    Float,
    Instance,
    Nullable,
    NumberSpec,
    Required,
    Seq,
    String,
    field,
)
from ..model import Model

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'CumSum',
    'CustomJSExpr',
    'Expression',
    'PolarTransform',
    'Stack',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class Expression(Model):
    ''' Base class for ``Expression`` models that represent a computation
    to be carried out on the client-side.

    JavaScript implementations should implement the following methods:

    .. code-block

        v_compute(source: ColumnarDataSource): Arrayable {
            # compute and return array of values
        }

    .. note::
        If you wish for results to be cached per source and updated only if
        the source changes, implement ``_v_compute: (source)`` instead.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class CustomJSExpr(Expression):
    ''' Evaluate a JavaScript function/generator.

    .. warning::
        The explicit purpose of this Bokeh Model is to embed *raw JavaScript
        code* for a browser to execute. If any part of the code is derived
        from untrusted user inputs, then you must take appropriate care to
        sanitize the user input prior to passing to Bokeh.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    args = Dict(String, AnyRef, help="""
    A mapping of names to Python objects. In particular those can be bokeh's models.
    These objects are made available to the callback's code snippet as the values of
    named parameters to the callback. There is no need to manually include the data
    source of the associated glyph renderer, as it is available within the scope of
    the code via `this` keyword (e.g. `this.data` will give access to raw data).
    """)

    code = String(default="", help="""
    A snippet of JavaScript code to execute in the browser. The code is made into
    the body of a generator function, and all of of the named objects in ``args``
    are available as parameters that the code can use. One can either return an
    array-like object (array, typed array, nd-array), an iterable (which will
    be converted to an array) or a scalar value (which will be converted into
    an array of an appropriate length), or alternatively yield values that will
    be collected into an array.
    """)


class CumSum(Expression):
    ''' An expression for generating arrays by cumulatively summing a single
    column from a ``ColumnDataSource``.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    field = Required(String, help="""
    The name of a ``ColumnDataSource`` column to cumulatively sum for new values.
    """)

    include_zero = Bool(default=False, help="""
    Whether to include zero at the start of the result. Note that the length
    of the result is always the same as the input column. Therefore if this
    property is True, then the last value of the column will not be included
    in the sum.

    .. code-block:: python

        source = ColumnDataSource(data=dict(foo=[1, 2, 3, 4]))

        CumSum(field='foo')
        # -> [1, 3, 6, 10]

        CumSum(field='foo', include_zero=True)
        # -> [0, 1, 3, 6]

    """)


class Stack(Expression):
    ''' An expression for generating arrays by summing different columns from
    a ``ColumnDataSource``.

    This expression is useful for implementing stacked bar charts at a low
    level.

    '''

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    fields = Seq(String, default=[], help="""
    A sequence of fields from a ``ColumnDataSource`` to sum (elementwise). For
    example:

    .. code-block:: python

        Stack(fields=['sales', 'marketing'])

    Will compute an array of values (in the browser) by adding the elements
    of the ``'sales'`` and ``'marketing'`` columns of a data source.
    """)


@abstract
class ScalarExpression(Model):
    """ Base class for for scalar expressions. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class Minimum(ScalarExpression):
    """ Computes minimum value of a data source's column. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    field = Required(String)
    initial = Nullable(Float, default=inf)


class Maximum(ScalarExpression):
    """ Computes maximum value of a data source's column. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    field = Required(String)
    initial = Nullable(Float, default=-inf)


@abstract
class CoordinateTransform(Expression):
    """ Base class for coordinate transforms. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @property
    def x(self):
        return XComponent(transform=self)

    @property
    def y(self):
        return YComponent(transform=self)


class PolarTransform(CoordinateTransform):
    """ Transform from polar to cartesian coordinates. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    radius = NumberSpec(default=field("radius"), help="""
    The radial coordinate (i.e. the distance from the origin).

    Negative radius is allowed, which is equivalent to using positive radius
    and changing ``direction`` to the opposite value.
    """)

    angle = AngleSpec(default=field("angle"), help="""
    The angular coordinate (i.e. the angle from the reference axis).
    """)

    direction = Enum(Direction, default=Direction.anticlock, help="""
    Whether ``angle`` measures clockwise or anti-clockwise from the reference axis.
    """)


@abstract
class XYComponent(Expression):
    """ Base class for bi-variate expressions. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    transform = Instance(CoordinateTransform)


class XComponent(XYComponent):
    """ X-component of a coordinate system transform to cartesian coordinates. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class YComponent(XYComponent):
    """ Y-component of a coordinate system transform to cartesian coordinates. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2023, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
""" Various abstractions over the CSS object model.

"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Bokeh imports
from ..core.has_props import abstract
from ..core.properties import Nullable, Required, String
from ..model import Model

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    "GlobalImportedStyleSheet",
    "GlobalInlineStyleSheet",
    "ImportedStyleSheet",
    "InlineStyleSheet",
    "Styles",
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

@abstract
class StyleSheet(Model):
    """ """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

class InlineStyleSheet(StyleSheet):
    """ Inline stylesheet equivalent to ``<style type="text/css">${css}</style>``.

    .. note::
        Depending on the context, this stylesheet will be appended either to
        the the parent shadow root, if used in a component, or otherwise to
        the ``<head>`` element. If you want to append globally regardless of
        the context, use ``GlobalInlineStyleSheet`` instead.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    css = Required(String, help="""
    The contents of this stylesheet.
    """)

class ImportedStyleSheet(StyleSheet):
    """ Imported stylesheet equivalent to ``<link rel="stylesheet" href="${url}">``.

    .. note::
        Depending on the context, this stylesheet will be appended either to
        the the parent shadow root, if used in a component, or otherwise to
        the ``<head>`` element. If you want to append globally regardless of
        the context, use ``GlobalImportedStyleSheet`` instead.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    url = Required(String, help="""
    The location of an external stylesheet.
    """)

class GlobalInlineStyleSheet(InlineStyleSheet):
    """ An inline stylesheet that's appended to the ``<head>`` element.

    .. note::
        A stylesheet will be appended only once, regardless of how
        many times it's being used in other models.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

class GlobalImportedStyleSheet(ImportedStyleSheet):
    """ An imported stylesheet that's appended to the ``<head>`` element.

    .. note::
        A stylesheet will be appended only once, regardless of how
        many times it's being used in other models.
    """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

class Styles(Model):
    """ Allows to configure style attribute of DOM elements. """

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    align_content = Nullable(String)
    align_items = Nullable(String)
    align_self = Nullable(String)
    alignment_baseline = Nullable(String)
    all = Nullable(String)
    animation = Nullable(String)
    animation_delay = Nullable(String)
    animation_direction = Nullable(String)
    animation_duration = Nullable(String)
    animation_fill_mode = Nullable(String)
    animation_iteration_count = Nullable(String)
    animation_name = Nullable(String)
    animation_play_state = Nullable(String)
    animation_timing_function = Nullable(String)
    aspect_ratio = Nullable(String)
    backface_visibility = Nullable(String)
    background = Nullable(String)
    background_attachment = Nullable(String)
    background_clip = Nullable(String)
    background_color = Nullable(String)
    background_image = Nullable(String)
    background_origin = Nullable(String)
    background_position = Nullable(String)
    background_position_x = Nullable(String)
    background_position_y = Nullable(String)
    background_repeat = Nullable(String)
    background_size = Nullable(String)
    baseline_shift = Nullable(String)
    block_size = Nullable(String)
    border = Nullable(String)
    border_block_end = Nullable(String)
    border_block_end_color = Nullable(String)
    border_block_end_style = Nullable(String)
    border_block_end_width = Nullable(String)
    border_block_start = Nullable(String)
    border_block_start_color = Nullable(String)
    border_block_start_style = Nullable(String)
    border_block_start_width = Nullable(String)
    border_bottom = Nullable(String)
    border_bottom_color = Nullable(String)
    border_bottom_left_radius = Nullable(String)
    border_bottom_right_radius = Nullable(String)
    border_bottom_style = Nullable(String)
    border_bottom_width = Nullable(String)
    border_collapse = Nullable(String)
    border_color = Nullable(String)
    border_image = Nullable(String)
    border_image_outset = Nullable(String)
    border_image_repeat = Nullable(String)
    border_image_slice = Nullable(String)
    border_image_source = Nullable(String)
    border_image_width = Nullable(String)
    border_inline_end = Nullable(String)
    border_inline_end_color = Nullable(String)
    border_inline_end_style = Nullable(String)
    border_inline_end_width = Nullable(String)
    border_inline_start = Nullable(String)
    border_inline_start_color = Nullable(String)
    border_inline_start_style = Nullable(String)
    border_inline_start_width = Nullable(String)
    border_left = Nullable(String)
    border_left_color = Nullable(String)
    border_left_style = Nullable(String)
    border_left_width = Nullable(String)
    border_radius = Nullable(String)
    border_right = Nullable(String)
    border_right_color = Nullable(String)
    border_right_style = Nullable(String)
    border_right_width = Nullable(String)
    border_spacing = Nullable(String)
    border_style = Nullable(String)
    border_top = Nullable(String)
    border_top_color = Nullable(String)
    border_top_left_radius = Nullable(String)
    border_top_right_radius = Nullable(String)
    border_top_style = Nullable(String)
    border_top_width = Nullable(String)
    border_width = Nullable(String)
    bottom = Nullable(String)
    box_shadow = Nullable(String)
    box_sizing = Nullable(String)
    break_after = Nullable(String)
    break_before = Nullable(String)
    break_inside = Nullable(String)
    caption_side = Nullable(String)
    caret_color = Nullable(String)
    clear = Nullable(String)
    clip = Nullable(String)
    clip_path = Nullable(String)
    clip_rule = Nullable(String)
    color = Nullable(String)
    color_interpolation = Nullable(String)
    color_interpolation_filters = Nullable(String)
    column_count = Nullable(String)
    column_fill = Nullable(String)
    column_gap = Nullable(String)
    column_rule = Nullable(String)
    column_rule_color = Nullable(String)
    column_rule_style = Nullable(String)
    column_rule_width = Nullable(String)
    column_span = Nullable(String)
    column_width = Nullable(String)
    columns = Nullable(String)
    content = Nullable(String)
    counter_increment = Nullable(String)
    counter_reset = Nullable(String)
    cursor = Nullable(String)
    direction = Nullable(String)
    display = Nullable(String)
    dominant_baseline = Nullable(String)
    empty_cells = Nullable(String)
    fill = Nullable(String)
    fill_opacity = Nullable(String)
    fill_rule = Nullable(String)
    filter = Nullable(String)
    flex = Nullable(String)
    flex_basis = Nullable(String)
    flex_direction = Nullable(String)
    flex_flow = Nullable(String)
    flex_grow = Nullable(String)
    flex_shrink = Nullable(String)
    flex_wrap = Nullable(String)
    float = Nullable(String)
    flood_color = Nullable(String)
    flood_opacity = Nullable(String)
    font = Nullable(String)
    font_family = Nullable(String)
    font_feature_settings = Nullable(String)
    font_kerning = Nullable(String)
    font_size = Nullable(String)
    font_size_adjust = Nullable(String)
    font_stretch = Nullable(String)
    font_style = Nullable(String)
    font_synthesis = Nullable(String)
    font_variant = Nullable(String)
    font_variant_caps = Nullable(String)
    font_variant_east_asian = Nullable(String)
    font_variant_ligatures = Nullable(String)
    font_variant_numeric = Nullable(String)
    font_variant_position = Nullable(String)
    font_weight = Nullable(String)
    gap = Nullable(String)
    glyph_orientation_vertical = Nullable(String)
    grid = Nullable(String)
    grid_area = Nullable(String)
    grid_auto_columns = Nullable(String)
    grid_auto_flow = Nullable(String)
    grid_auto_rows = Nullable(String)
    grid_column = Nullable(String)
    grid_column_end = Nullable(String)
    grid_column_gap = Nullable(String)
    grid_column_start = Nullable(String)
    grid_gap = Nullable(String)
    grid_row = Nullable(String)
    grid_row_end = Nullable(String)
    grid_row_gap = Nullable(String)
    grid_row_start = Nullable(String)
    grid_template = Nullable(String)
    grid_template_areas = Nullable(String)
    grid_template_columns = Nullable(String)
    grid_template_rows = Nullable(String)
    height = Nullable(String)
    hyphens = Nullable(String)
    image_orientation = Nullable(String)
    image_rendering = Nullable(String)
    inline_size = Nullable(String)
    justify_content = Nullable(String)
    justify_items = Nullable(String)
    justify_self = Nullable(String)
    left = Nullable(String)
    letter_spacing = Nullable(String)
    lighting_color = Nullable(String)
    line_break = Nullable(String)
    line_height = Nullable(String)
    list_style = Nullable(String)
    list_style_image = Nullable(String)
    list_style_position = Nullable(String)
    list_style_type = Nullable(String)
    margin = Nullable(String)
    margin_block_end = Nullable(String)
    margin_block_start = Nullable(String)
    margin_bottom = Nullable(String)
    margin_inline_end = Nullable(String)
    margin_inline_start = Nullable(String)
    margin_left = Nullable(String)
    margin_right = Nullable(String)
    margin_top = Nullable(String)
    marker = Nullable(String)
    marker_end = Nullable(String)
    marker_mid = Nullable(String)
    marker_start = Nullable(String)
    mask = Nullable(String)
    mask_composite = Nullable(String)
    mask_image = Nullable(String)
    mask_position = Nullable(String)
    mask_repeat = Nullable(String)
    mask_size = Nullable(String)
    mask_type = Nullable(String)
    max_block_size = Nullable(String)
    max_height = Nullable(String)
    max_inline_size = Nullable(String)
    max_width = Nullable(String)
    min_block_size = Nullable(String)
    min_height = Nullable(String)
    min_inline_size = Nullable(String, help="""
    The `min-inline-size`_ CSS property defines the horizontal or vertical
    minimal size of an element's block, depending on its writing mode. It
    corresponds to either the ``min-width`` or the ``min-height`` property,
    depending on the value of ``writing-mode``.

    .. _min-inline-size: https://developer.mozilla.org/en-US/docs/Web/CSS/min-inline-size
    """)

    min_width = Nullable(String)
    object_fit = Nullable(String)
    object_position = Nullable(String)
    opacity = Nullable(String)
    order = Nullable(String)
    orphans = Nullable(String)
    outline = Nullable(String)
    outline_color = Nullable(String)
    outline_offset = Nullable(String)
    outline_style = Nullable(String)
    outline_width = Nullable(String)
    overflow = Nullable(String)
    overflow_anchor = Nullable(String)
    overflow_wrap = Nullable(String)
    overflow_x = Nullable(String)
    overflow_y = Nullable(String)
    overscroll_behavior = Nullable(String)
    overscroll_behavior_block = Nullable(String)
    overscroll_behavior_inline = Nullable(String)
    overscroll_behavior_x = Nullable(String)
    overscroll_behavior_y = Nullable(String)
    padding = Nullable(String)
    padding_block_end = Nullable(String)
    padding_block_start = Nullable(String)
    padding_bottom = Nullable(String)
    padding_inline_end = Nullable(String)
    padding_inline_start = Nullable(String)
    padding_left = Nullable(String)
    padding_right = Nullable(String)
    padding_top = Nullable(String)
    page_break_after = Nullable(String)
    page_break_before = Nullable(String)
    page_break_inside = Nullable(String)
    paint_order = Nullable(String)
    perspective = Nullable(String)
    perspective_origin = Nullable(String)
    place_content = Nullable(String)
    place_items = Nullable(String)
    place_self = Nullable(String)
    pointer_events = Nullable(String)
    position = Nullable(String)
    quotes = Nullable(String)
    resize = Nullable(String)
    right = Nullable(String)
    rotate = Nullable(String)
    row_gap = Nullable(String)
    ruby_align = Nullable(String)
    ruby_position = Nullable(String)
    scale = Nullable(String)
    scroll_behavior = Nullable(String)
    shape_rendering = Nullable(String)
    stop_color = Nullable(String)
    stop_opacity = Nullable(String)
    stroke = Nullable(String)
    stroke_dasharray = Nullable(String)
    stroke_dashoffset = Nullable(String)
    stroke_linecap = Nullable(String)
    stroke_linejoin = Nullable(String)
    stroke_miterlimit = Nullable(String)
    stroke_opacity = Nullable(String)
    stroke_width = Nullable(String)
    tab_size = Nullable(String)
    table_layout = Nullable(String)
    text_align = Nullable(String)
    text_align_last = Nullable(String)
    text_anchor = Nullable(String)
    text_combine_upright = Nullable(String)
    text_decoration = Nullable(String)
    text_decoration_color = Nullable(String)
    text_decoration_line = Nullable(String)
    text_decoration_style = Nullable(String)
    text_emphasis = Nullable(String)
    text_emphasis_color = Nullable(String)
    text_emphasis_position = Nullable(String)
    text_emphasis_style = Nullable(String)
    text_indent = Nullable(String)
    text_justify = Nullable(String)
    text_orientation = Nullable(String)
    text_overflow = Nullable(String)
    text_rendering = Nullable(String)
    text_shadow = Nullable(String)
    text_transform = Nullable(String)
    text_underline_position = Nullable(String)
    top = Nullable(String)
    touch_action = Nullable(String)
    transform = Nullable(String)
    transform_box = Nullable(String)
    transform_origin = Nullable(String)
    transform_style = Nullable(String)
    transition = Nullable(String)
    transition_delay = Nullable(String)
    transition_duration = Nullable(String)
    transition_property = Nullable(String)
    transition_timing_function = Nullable(String)
    translate = Nullable(String)
    unicode_bidi = Nullable(String)
    user_select = Nullable(String)
    vertical_align = Nullable(String)
    visibility = Nullable(String)
    white_space = Nullable(String)
    widows = Nullable(String)
    width = Nullable(String)
    will_change = Nullable(String)
    word_break = Nullable(String)
    word_spacing = Nullable(String)
    word_wrap = Nullable(String)
    writing_mode = Nullable(String)
    z_index = Nullable(String)

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------
